--[[
    Aaron MacDonald
    August 18, 2010
]]

require('graphics')
require('constants')
require('actor')
require('levels')
require('ai')

local FPS = 60
local SECONDS_PER_FRAME = 1 / FPS

local accumulatedTime

local isPaused

----------------------------------------------------------------------

local KEY_QUIT = 'escape'

local VALID_BUTTONS = { [constants.UP] = true,
                        [constants.DOWN] = true,
                        [constants.LEFT] = true,
                        [constants.RIGHT] = true }

----------------------------------------------------------------------

local PLAYER_TYPE = {
    spriteSheet = 'player',
    spriteWidth = 64, spriteHeight = 96,
    spriteOffsetX = 32, spriteOffsetY = 64,
    radius = 32,
    speed = 600, diagonalSpeed = actor.diagonalSpeed(600)
}

local NPC_TYPE = {
    spriteSheet = 'npc',
    spriteWidth = 64, spriteHeight = 96,
    spriteOffsetX = 32, spriteOffsetY = 64,
    radius = 32,
    speed = 600, diagonalSpeed = actor.diagonalSpeed(600)
}

local PLAYER_START_FACING = constants.DOWN

----------------------------------------------------------------------

local currentButtons = { [constants.UP] = false,
                         [constants.DOWN] = false,
                         [constants.LEFT] = false,
                         [constants.RIGHT] = false }

local currentLevel
local allActors
local player

local camera

local world

--local backgroundMusic

----------------------------------------------------------------------

local wanderAISteering = {
    [constants.NORTH] =
        function(movement)
            movement[constants.DOWN] = false
            movement[constants.LEFT] = false
            movement[constants.RIGHT] = false

            movement[constants.UP] = true
        end,
    [constants.NORTHEAST] =
        function(movement)
            movement[constants.DOWN] = false
            movement[constants.LEFT] = false

            movement[constants.UP] = true
            movement[constants.RIGHT] = true
        end,
    [constants.EAST] =
        function(movement)
            movement[constants.UP] = false
            movement[constants.DOWN] = false
            movement[constants.LEFT] = false

            movement[constants.RIGHT] = true
        end,
    [constants.SOUTHEAST] =
        function(movement)
            movement[constants.UP] = false
            movement[constants.LEFT] = false

            movement[constants.DOWN] = true
            movement[constants.RIGHT] = true
        end,
    [constants.SOUTH] =
        function(movement)
            movement[constants.UP] = false
            movement[constants.LEFT] = false
            movement[constants.RIGHT] = false

            movement[constants.DOWN] = true
        end,
    [constants.SOUTHWEST] =
        function(movement)
            movement[constants.UP] = false
            movement[constants.RIGHT] = false

            movement[constants.DOWN] = true
            movement[constants.LEFT] = true
        end,
    [constants.WEST] =
        function(movement)
            movement[constants.UP] = false
            movement[constants.DOWN] = false
            movement[constants.RIGHT] = false

            movement[constants.LEFT] = true
        end,
    [constants.NORTHWEST] =
        function(movement)
            movement[constants.DOWN] = false
            movement[constants.RIGHT] = false

            movement[constants.UP] = true
            movement[constants.LEFT] = true
        end,
}

local function wanderAISelectDir(brain)
    local i = math.random(1, #brain.possibleDirs)
    dir = brain.possibleDirs[i]
    wanderAISteering[dir](brain.movement)
end

local function wanderAIInit(brain)
    brain.possibleDirs = {constants.NORTH, constants.NORTHEAST,
      constants.EAST, constants.SOUTHEAST, constants.SOUTH,
      constants.SOUTHWEST, constants.WEST, constants.NORTHWEST}

    wanderAISelectDir(brain)

    brain.travelTime = 10
end

local function wanderAIThink(brain)
    if brain.travelTime == 0 then
        wanderAISelectDir(brain)
        brain.travelTime = 10
    else
        brain.travelTime = brain.travelTime - 1
    end
end

local function wanderHandleWallHit(brain)
    wanderAISelectDir(brain)
    brain.travelTime = 10
end

----------------------------------------------------------------------

local function spawnNPCs(world, level)
    local newNPCs = {}

    if level.npcSpawn then
        for _, data in ipairs(level.npcSpawn) do
            local npc = actor.newActor(world, NPC_TYPE,
              data.x, data.y, constants.DOWN)

            npc.ai = ai.newBrain(5, wanderAIThink, wanderAIInit,
              wanderHandleWallHit)

            table.insert(newNPCs, npc)
        end
    end

    return newNPCs
end

----------------------------------------------------------------------

function love.load()
    math.randomseed( os.time() )

    accumulatedTime = 0
    isPaused = false

    graphics.loadFont()

    currentLevel = levels.newLevel('test_level')

    -- Physics

    world = love.physics.newWorld(levels.pixelWidth(currentLevel),
      levels.truePixelHeight(currentLevel))
    world:setGravity(0, 0)
    world:setMeter(levels.TILE_SIZE)

    walls = {}

    for x = 0, currentLevel.width - 1 do
        for y = 0, currentLevel.height - 1 do
            if levels.isWall(currentLevel, x, y) then
                local tx = x * levels.TILE_SIZE
                local ty = y * levels.TILE_SIZE

                local wallBody = love.physics.newBody(world,
                  tx + levels.TILE_HALF_SIZE,
                  ty + levels.TILE_HALF_SIZE,
                  0, 0)
                local wallShape = love.physics.newRectangleShape(
                  wallBody, 0, 0, levels.TILE_SIZE, levels.TILE_SIZE,
                  0 )
                wallShape:setRestitution(0)
                wallShape:setData(constants.WALL)

                table.insert(walls,
                  {body = wallBody, shape = wallShape})
            end
        end
    end

    --

    local startX = 0
    local startY = 0

    if currentLevel.playerSpawn then
        startX = currentLevel.playerSpawn.x
        startY = currentLevel.playerSpawn.y
    end

    allActors = spawnNPCs(world, currentLevel)

    player = actor.newActor(world, PLAYER_TYPE, startX, startY,
      PLAYER_START_FACING)

    table.insert(allActors, player)

    local width = love.graphics.getWidth()
    local height = love.graphics.getHeight()

    camera = {
        x = 0,
        y = 0,
        minX = width / 2,
        minY = height / 2,
        maxX = levels.pixelWidth(currentLevel) - width,
        maxY = levels.pixelHeight(currentLevel) - height,
        width = width,
        height = height
    }
end

----------------------------------------------------------------------

function love.keypressed( key, unicode )
    if key == KEY_QUIT then
        love.event.push('q')
    elseif VALID_BUTTONS[key] then
        currentButtons[key] = true
    end
end

function love.keyreleased( key, unicode )
   if VALID_BUTTONS[key] then
       currentButtons[key] = false
   end
end

function love.focus(hasFocus)
    isPaused = not hasFocus
end

----------------------------------------------------------------------

local function updateCamera()
    if actor.getX(player) < camera.minX then
        camera.x = 0
    elseif actor.getX(player) > (camera.maxX + camera.minX) then
        camera.x = camera.maxX
    else
        camera.x = actor.getX(player) - camera.minX
    end

    if (actor.getY(player) - currentLevel.minY) < camera.minY then
        camera.y = currentLevel.minY
    elseif actor.getY(player) > (camera.maxY + camera.minY) then
        camera.y = camera.maxY
    else
        camera.y = actor.getY(player) - camera.minY
    end
end

local function myUpdate()
    -- Update velocities
    for _, act in ipairs(allActors) do
        local movementControls
        if act == player then
            movementControls = currentButtons
        else
            ai.think(act.ai)
            movementControls = act.ai.movement
        end

        actor.controlMovement(act, movementControls)

        if act ~= player then
            ai.update(act.ai)
        end
    end

    world:update(SECONDS_PER_FRAME)

    updateCamera()
end

function love.update(dt)
    accumulatedTime = accumulatedTime + dt
    while accumulatedTime >= SECONDS_PER_FRAME
    do
        if not isPaused then
            myUpdate()
        end
        accumulatedTime = accumulatedTime - SECONDS_PER_FRAME
    end
end

----------------------------------------------------------------------

local function spriteCompare(a, b)
    local ay = actor.getY(a) + (a.type.spriteHeight - a.type.spriteOffsetY)
    local by = actor.getY(b) + (b.type.spriteHeight - b.type.spriteOffsetY)

    return ay < by
end

function love.draw()
    love.graphics.translate(-camera.x, -camera.y)

    levels.drawGround(currentLevel, camera)

    table.sort(allActors, spriteCompare)

    for _, act in ipairs(allActors) do
        actor.draw(act)
    end

    levels.drawOver(currentLevel, camera)

    love.graphics.translate(camera.x, camera.y)

    love.graphics.print(actor.getX(player) .. ', '
      .. actor.getY(player), 0, 0)
end